"""
Operational semantics for the Why language

Provides evaluation functions for the three syntactic classes,
based on a environment mapping variables to integer values.
These functions return None when an error occur somewhere (Big step semantics)
"""

from syntax import *

def eval_aexp(env: dict[str,int], exp: ArithExpr) -> int | None:
    """evaluates an arithmetic expression"""
    match exp:
        case AECst(value):
            return value
        case AEVar(var):
            if var in env:
                return env[var]
            else:
                return None
        case AEUop(uop,expr):
            val = eval_aexp(env, expr)
            if val is None:
                return None
            
            # Match on the unary operator
            if uop == Uop.OPP:
                return -val
            else:
                return None
        case AEBop(bop,left_expr,right_expr):
            left_val = eval_aexp(env, left_expr)
            right_val = eval_aexp(env, right_expr)
            if left_val is None or right_val is None:
                return None
            
            if bop == Bop.ADD:
                return left_val + right_val
            elif bop == Bop.MUL:
                return left_val * right_val
            # check for other potential operators in Bop
            elif str(bop) == "Bop.SUB" or str(bop) == "-":
                return left_val - right_val
            elif str(bop) == "Bop.DIV" or str(bop) == "/":
                if right_val == 0:
                    return None  # div by zero
                return left_val // right_val
            else:
                return None
        case _: assert False

def eval_bexp(env: dict[str,int], exp:BoolExpr) -> bool | None:
    """evaluates a boolean expression"""
    match exp:
        case BEPlain(aexpr):
            val = eval_aexp(env, aexpr)
            if val is None:
                return None
            return val != 0
        case BEEq(left_expr,right_expr):
            left_val = eval_aexp(env, left_expr)
            right_val = eval_aexp(env, right_expr)
            if left_val is None or right_val is None:
                return None
            return left_val == right_val
        case BELeq(left_expr,right_expr):
            left_val = eval_aexp(env, left_expr)
            right_val = eval_aexp(env, right_expr)
            if left_val is None or right_val is None:
                return None
            return left_val <= right_val
        case BENeg(expr):
            val = eval_bexp(env, expr)
            if val is None:
                return None
            return not val
        case _: assert False

def eval_instr(env: dict[str,int], instr: Instr) -> dict[str,int] | None:
    """evaluates an instruction"""
    match instr:
        case ISkip():
            return env.copy()
        case ISet(var,expr):
            val = eval_aexp(env, expr)
            if val is None:
                return None
            new_env = env.copy()
            new_env[var] = val
            return new_env
        case ISeq(first,second):
            env_after_first = eval_instr(env, first)
            if env_after_first is None:
                return None
            return eval_instr(env_after_first, second)
        case ICond(cond,true_branch,false_branch):
            cond_val = eval_bexp(env, cond)
            if cond_val is None:
                return None
            if cond_val:
                return eval_instr(env, true_branch)
            else:
                return eval_instr(env, false_branch)
        case ILoop(cond,body):
            cond_val = eval_bexp(env, cond)
            if cond_val is None:
                return None
            if not cond_val:
                return env.copy()
            env_after_body = eval_instr(env, body)
            if env_after_body is None:
                return None
            return eval_instr(env_after_body, ILoop(cond, body))
        case _: assert False
